#!/usr/bin/env python
# vim: et ai ts=4 sw=4:

from charmhelpers.contrib.openstack.utils import configure_installation_source
from charmhelpers.core import hookenv
from charmhelpers.core.hookenv import ERROR, INFO

import json
import os
import sys
from util import StorageServiceUtil, generate_volume_label

hooks = hookenv.Hooks()

PERSIST_DATA_FILE = "charm-data"


# XXX We can remove _persist_data when juju bug:1279018 is fixed to provide
# relation-data from *-relation-departed hook runs
def _persist_data(key, value):
    """Save C{key} and C{value} pairs in a persistent file for reference"""
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    if os.path.exists(filepath):
        with open(filepath) as inputfile:
            data = json.load(inputfile)
    data[key] = value
    with open(filepath, "w") as outfile:
        outfile.write(unicode(json.dumps(data, ensure_ascii=False)))


def _get_persistent_data(key=None, remove_values=False):
    """Get a persistent data as specified by C{key} from the PERSIST_DATA_FILE

    If specified C{key} does not exist, or if no PERSIST_DATA_FILE exists,
    return C{None}. If C{remove_values} is C{True} remove the specified C{key}
    from the data file.
    """
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    if os.path.exists(filepath):
        with open(filepath) as inputfile:
            data = json.load(inputfile)
    if key:
        value = data.get(key, None)
        if remove_values:
            data.pop(key, None)
            if data:
                with open(filepath, "w") as outfile:
                    outfile.write(
                        unicode(json.dumps(data, ensure_ascii=False)))
            elif os.path.exists(filepath):
                os.remove(filepath)
    else:
        value = data
        if not data:
            value = None
        if remove_values:
            if os.path.exists(filepath):
                os.remove(filepath)
    return value


@hooks.hook()
def config_changed():
    """Update changed endpoints and credentials"""
    storage_util = StorageServiceUtil(hookenv.config("provider"))
    missing_options = []
    for config_option in storage_util.required_config_options:
        if not hookenv.config(config_option):
            missing_options.append(config_option)
    if missing_options:
        hookenv.log(
            "WARNING: Incomplete charm config. Missing values for: %s" %
            ",".join(missing_options), hookenv.WARNING)
        sys.exit(0)
    storage_util.load_environment()


@hooks.hook()
def install():
    """Install required packages if not present"""
    from charmhelpers import fetch
    provider = hookenv.config("provider")
    if provider == "nova":
        configure_installation_source(hookenv.config('openstack-origin'))
        required_packages = ["python-novaclient"]
    elif provider == "ec2":
        required_packages = ["euca2ools"]
    fetch.apt_update(fatal=True)
    fetch.apt_install(required_packages, fatal=True)


@hooks.hook("block-storage-relation-departed")
def block_storage_relation_departed():
    """Detach a volume from its related instance when relation departs"""
    storage_util = StorageServiceUtil(hookenv.config("provider"))

    # XXX juju bug:1279018 for departed-hooks to see relation-data
    volume_label = _get_persistent_data(
        key=hookenv.remote_unit(), remove_values=True)
    if not volume_label:
        hookenv.log(
            "Cannot detach volume from instance without volume_label",
            ERROR)
        sys.exit(1)
    storage_util.detach_volume(volume_label)


@hooks.hook("block-storage-relation-changed")
def block_storage_relation_changed():
    """Attach a volume to the C{instance-id} requested by the relation

    Optionally the relation can specify:
        - C{volume-label} the label to set on the created volume
        - C{volume-id} to attach an existing volume-id or volume-name
        - C{size} to request a specific volume size
    """
    storage_util = StorageServiceUtil(hookenv.config("provider"))
    instance_id = hookenv.relation_get('instance-id')
    volume_label = hookenv.relation_get('volume-label')
    if not instance_id:
        hookenv.log("Waiting for relation to define instance-id", INFO)
        return
    volume_id = hookenv.relation_get('volume-id')  # optional
    size = hookenv.relation_get('size')            # optional
    device_path = storage_util.attach_volume(
        instance_id=instance_id, volume_id=volume_id, size=size,
        volume_label=volume_label)
    remote_unit = hookenv.remote_unit()
    if not volume_label:
        volume_label = generate_volume_label(remote_unit)
    _persist_data(remote_unit, volume_label)

    # Volume is attached, send the path back to the remote storage unit
    hookenv.relation_set(relation_settings={"block-device-path": device_path})


if __name__ == '__main__':
    hook_name = os.path.basename(sys.argv[0])
    hookenv.log("Running {} hook".format(hook_name))
    if hookenv.relation_id():
        hookenv.log("Relation {} with {}".format(
            hookenv.relation_id(), hookenv.remote_unit()))
    hooks.execute(sys.argv)
