from charmhelpers import fetch
from charmhelpers.contrib.openstack.utils import configure_installation_source
import hooks
import json
import mocker
import os
from testing import TestHookenv
from util import StorageServiceUtil


class TestHooks(mocker.MockerTestCase):

    def setUp(self):
        super(TestHooks, self).setUp()
        self.maxDiff = None
        hooks.hookenv = TestHookenv(
            {"key": "myusername", "tenant": "myusername_project",
             "secret": "password", "region": "region1", "provider": "nova",
             "endpoint": "https://keystone_url:443/v2.0/",
             "default_volume_size": 11,
             "openstack-origin": "cloud:precise-folsom/staging"})

    def test_wb_persist_data_creates_persist_file_if_it_doesnt_exist(self):
        """
        L{_persist_data} will create a new PERSIST_DATA_FILE in C{CHARM_DIR}
        if the file does not already exist and store the provided C{key},
        C{value} pairs.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        self.assertFalse(os.path.exists(persist_path))
        hooks._persist_data("key", "value")
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["key"], "value")

    def test_wb_persist_data_appends_new_values_to_existing_persist_file(self):
        """
        L{_persist_data} will add new values to an existing PERSIST_DATA_FILE
        in C{CHARM_DIR} without removing existing C{key}, C{value} pairs.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "value1"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        hooks._persist_data("key2", "value2")
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["key1"], "value1")
            self.assertEqual(data["key2"], "value2")

    def test_wb_get_persistent_data_returns_none_when_no_persist_file(self):
        """
        L{_get_persistent_data} will return C{None} for any C{key} search if
        the PERSIST_DATA_FILE does not yet exist.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        self.assertFalse(os.path.exists(persist_path))
        self.assertIsNone(hooks._get_persistent_data("key2"))
        self.assertIsNone(hooks._get_persistent_data())

    def test_wb_get_persistent_data_returns_none_when_no_key_exists(self):
        """
        L{_get_persistent_data} will return C{None} for any C{key} search when
        the C{key} is not defined in the PERSIST_DATA_FILE.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertIsNone(hooks._get_persistent_data("cant-find-this-key"))

    def test_wb_get_persistent_data_returns_value_when_key_exists(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(hooks._get_persistent_data("key1"), "i-123")

    def test_wb_get_persistent_data_removes_key_when_remove_values_true(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE. It will also remove that key from the
        PERSIST_DATA_FILE when C{remove_values} is C{True}.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "i-123", "key2": "i-456"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            hooks._get_persistent_data("key1", remove_values=True), "i-123")
        with open(persist_path) as infile:
            new_data = json.load(infile)
            self.assertEqual(new_data, {"key2": "i-456"})

    def test_wb_get_persistent_data_removes_file_when_remove_values_true(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE. It will also remove the PERSIST_DATA_FILE when
        C{remove_values} is C{True} and the last remaining C{key} has been
        removed.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            hooks._get_persistent_data("key1", remove_values=True), "i-123")
        self.assertFalse(os.path.exists(persist_path))

    def test_wb_get_persistent_data_removes_file_when_no_key_specified(self):
        """
        L{_get_persistent_data} will return all existing values and remove the
        PERSIST_DATA_FILE when C{remove_values} is C{True} and no C{key} is
        specified.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": charm_dir}
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"key1": "i-123", "key2": "i-456"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            hooks._get_persistent_data(key=None, remove_values=True),
            data)
        self.assertFalse(os.path.exists(persist_path))

    def test_config_changed_missing_required_config_option(self):
        """
        L{config_changed} will exit in error and log a message if required
        configuration options are missing.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_config", hooks.hookenv._config)
        hooks.hookenv._config = (
            ("key", "myusername"), ("tenant", ""), ("provider", "nova"),
            ("secret", "password"), ("region", ""),
            ("endpoint", "https://keystone_url:443/v2.0/"))

        result = self.assertRaises(SystemExit, hooks.config_changed)
        self.assertEqual(result.code, 0)
        message = (
            "WARNING: Incomplete charm config. Missing values for: "
            "region,tenant")
        self.assertIn(
            message, hooks.hookenv._log_WARNING, "Not logged- %s" % message)

    def test_config_changed_calls_load_environment_when_config_is_ready(self):
        """
        L{config_changed} will call L{load_environment} to load and validate
        configured credentials when all mandatory configuration options are
        set.
        """
        self.storage_util = self.mocker.patch(StorageServiceUtil)
        self.storage_util.load_environment()
        self.mocker.replay()
        hooks.config_changed()

    def test_install_installs_novaclient_from_openstack_origin_config(self):
        """
        L{install} will call the charmhelper's C{configure_installation_source}
        to add the configured C{openstack-origin} the install appropriate
        version of the C{python-novaclient} package.
        """
        add_source = self.mocker.replace(configure_installation_source)
        add_source("cloud:precise-folsom/staging")
        apt_update = self.mocker.replace(fetch.apt_update)
        apt_update(fatal=True)
        apt_install = self.mocker.replace(fetch.apt_install)
        apt_install(["python-novaclient"], fatal=True)
        self.mocker.replay()

        hooks.install()

    def test_block_storage_relation_changed_waits_without_instance_id(self):
        """
        L{block_storage_relation_changed} will log a message and wait when
        the relation does not yet contain the required C{instance-id}.
        """
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = ()

        self.assertIsNone(hooks.hookenv.relation_get("instance-id"))
        hooks.block_storage_relation_changed()
        message = "Waiting for relation to define instance-id"
        self.assertIn(
            message, hooks.hookenv._log_INFO, "Not logged- %s" % message)

    def test_block_storage_relation_changed_with_instance_id(self):
        """
        L{block_storage_relation_changed} calls L{attach_volume} when
        C{instance-id} is available in the relation data. To report
        a successful device attach, it sets the relation data
        C{block-device-path} to the attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": "storage/0"}
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (("instance-id", "i-123"),)

        persist = self.mocker.replace(hooks._persist_data)
        persist("storage/0", "storage/0 unit volume")
        self.storage_util = self.mocker.patch(StorageServiceUtil)
        self.storage_util.attach_volume(
            instance_id="i-123", volume_id=None, size=None, volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertIsNone(hooks.hookenv.relation_get("volume-id"))
        self.assertIsNone(hooks.hookenv.relation_get("size"))
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_changed_with_instance_id_volume_id(self):
        """
        When C{volume-id} and C{instance-id} are both present in the relation
        data, they will both be passed to L{attach_volume}. To report a
        successful device attach, it sets the relation data
        C{block-device-path} to the attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        volume_id = "123-123-123"
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": "storage/0"}
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (
            ("instance-id", "i-123"), ("volume-id", volume_id))

        persist = self.mocker.replace(hooks._persist_data)
        persist("storage/0", "storage/0 unit volume")
        self.storage_util = self.mocker.patch(StorageServiceUtil)
        self.storage_util.attach_volume(
            instance_id="i-123", volume_id=volume_id, size=None,
            volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertEqual(hooks.hookenv.relation_get("volume-id"), volume_id)
        self.assertIsNone(hooks.hookenv.relation_get("size"))
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_changed_with_instance_id_size(self):
        """
        When C{size} and C{instance-id} are both present in the relation data,
        they will be passed to L{attach_volume}. To report a successful
        device attach, it sets the relation data C{block-device-path} to the
        attached volume's device path from nova.
        """
        device_path = "/dev/vdc"
        size = 12
        self.addCleanup(setattr, os, "environ", os.environ)
        os.environ = {"JUJU_REMOTE_UNIT": "storage/0"}
        self.addCleanup(
            setattr, hooks.hookenv, "_incoming_relation_data",
            hooks.hookenv._incoming_relation_data)
        hooks.hookenv._incoming_relation_data = (
            ("instance-id", "i-123"), ("size", size))
        persist = self.mocker.replace(hooks._persist_data)
        persist("storage/0", "storage/0 unit volume")
        self.storage_util = self.mocker.patch(StorageServiceUtil)
        self.storage_util.attach_volume(
            instance_id="i-123", volume_id=None, size=size, volume_label=None)
        self.mocker.result(device_path)   # The attached device path from nova
        self.mocker.replay()

        self.assertEqual(hooks.hookenv.relation_get("instance-id"), "i-123")
        self.assertIsNone(hooks.hookenv.relation_get("volume-id"))
        self.assertEqual(hooks.hookenv.relation_get("size"), size)
        self.assertIsNone(hooks.hookenv.relation_get("volume-label"))
        hooks.block_storage_relation_changed()
        self.assertEqual(
            hooks.hookenv._outgoing_relation_data,
            (("block-device-path", device_path),))

    def test_block_storage_relation_departed_fails_without_instance_id(self):
        """
        L{block_storage_relation_departed} will exit in error and logs a
        message when the persistent data does not contain the required
        C{instance-id} value for the C{hookenv.remote_unit}.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"JUJU_REMOTE_UNIT": "storage/0", "CHARM_DIR": charm_dir}

        result = self.assertRaises(
            SystemExit, hooks.block_storage_relation_departed)
        self.assertEqual(result.code, 1)
        message = "Cannot detach volume from instance without volume_label"
        self.assertIn(
            message, hooks.hookenv._log_ERROR, "Not logged- %s" % message)

    def test_block_storage_relation_departed_detaches_volume(self):
        """
        L{block_storage_relation_departed} calls L{detach_nova_volume} when
        C{instance-id} is available in the relation data.
        """
        self.addCleanup(setattr, os, "environ", os.environ)
        charm_dir = self.makeDir()
        os.environ = {"JUJU_REMOTE_UNIT": "storage/0", "CHARM_DIR": charm_dir}
        # Store our persistent instance-id info
        persist_path = "%s/%s" % (charm_dir, hooks.PERSIST_DATA_FILE)
        data = {"storage/0": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))

        self.storage_util = self.mocker.patch(StorageServiceUtil)
        self.storage_util.detach_volume("i-123")
        self.mocker.replay()

        hooks.block_storage_relation_departed()
