"""Common python utilities for the ec2 provider"""

from charmhelpers.core import hookenv
import subprocess
import os
import sys
from time import sleep

ENVIRONMENT_MAP = {
    "ec2": {"endpoint": "EC2_URL", "key": "EC2_ACCESS_KEY",
            "secret": "EC2_SECRET_KEY"},
    "nova": {"endpoint": "OS_AUTH_URL", "region": "OS_REGION_NAME",
             "tenant": "OS_TENANT_NAME", "key": "OS_USERNAME",
             "secret": "OS_PASSWORD"}}

REQUIRED_CONFIG_OPTIONS = {
    "ec2": ["endpoint", "key", "secret"],
    "nova": ["endpoint", "region", "tenant", "key", "secret"]}

PROVIDER_COMMANDS = {
    "ec2": {"validate": "euca-describe-instances",
            "detach": "euca-detach-volume -i %s %s"},
    "nova": {"validate": "nova list",
             "detach": "nova volume-detach %s %s"}}


class StorageServiceUtil(object):
    """Interact with an underlying cloud storage provider.
    Create, attach, label and detach storage volumes using EC2 or nova APIs.
    """
    provider = None
    environment_map = None
    required_config_options = None
    commands = None

    def __init__(self, provider):
        self.provider = provider
        if provider not in ENVIRONMENT_MAP:
            hookenv.log(
                "ERROR: Invalid charm configuration setting for provider. "
                "'%s' must be one of: %s" %
                (provider, ", ".join(ENVIRONMENT_MAP.keys())),
                hookenv.ERROR)
            sys.exit(1)
        self.environment_map = ENVIRONMENT_MAP[provider]
        self.commands = PROVIDER_COMMANDS[provider]
        self.required_config_options = REQUIRED_CONFIG_OPTIONS[provider]
        if provider == "ec2":
            import euca2ools.commands.euca.describevolumes as getvolumes
            import euca2ools.commands.euca.describeinstances as getinstances
            self.ec2_volume_class = getvolumes.DescribeVolumes
            self.ec2_instance_class = getinstances.DescribeInstances

    def load_environment(self):
        """
        Source our credentials from the configuration definitions into our
        operating environment
        """
        config_data = hookenv.config()
        for option in self.required_config_options:
            environment_variable = self.environment_map[option]
            os.environ[environment_variable] = config_data[option].strip()
        self.validate_credentials()

    def validate_credentials(self):
        """
        Attempt to contact the respective ec2 or nova volume service or exit(1)
        """
        try:
            subprocess.check_call(self.commands["validate"], shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log(
                "ERROR: Charm configured credentials can't access endpoint. "
                "%s" % str(e),
                hookenv.ERROR)
            sys.exit(1)
        hookenv.log(
            "Validated charm configuration credentials have access to block "
            "storage service")

    def describe_volumes(self, volume_id=None):
        method = getattr(self, "_%s_describe_volumes" % self.provider)
        return method(volume_id)

    def get_volume_id(self, volume_designation=None):
        """Return the ec2 or nova volume id associated with this unit

        Optionally, C{volume_designation} can be either a volume-id or
        volume-display-name and the matching C{volume-id} will be returned.
        If no matching volume is return C{None}.
        """
        matches = []
        volumes = self.describe_volumes()
        if volume_designation:
            token = volume_designation
        else:
            # Try to find volume label containing remote_unit name
            token = hookenv.remote_unit()
        for volume_id in volumes.keys():
            volume = volumes[volume_id]
            # Get volume by name or volume-id
            volume_name = volume["tags"].get("volume_name", "")
            if token == volume_id:
                matches.append(volume_id)
            elif token in volume_name:
                matches.append(volume_id)
        if len(matches) > 1:
            hookenv.log(
                "Error: Multiple volumes are associated with "
                "%s. Cannot get_volume_id." % token, hookenv.ERROR)
            sys.exit(1)
        elif matches:
            return matches[0]
        return None

    def attach_volume(self, instance_id, volume_id=None, size=None,
                      volume_label=None):
        """
        Create and attach a volume to the remote unit if none exists.

        Attempt to attach and validate the attached volume 10
        times. If unable to resolve the attach issues, exit in error and log
        the error.
        Log errors if the volume is in an unsupported state, and if C{in-use}
        report it is already attached.

        Return the device-path of the attached volume to the caller.
        """
        self.load_environment()   # Will fail if invalid environment
        remote_unit = hookenv.remote_unit()
        if volume_label is None:
            volume_label = generate_volume_label(remote_unit)
        if volume_id:
            volume = self.describe_volumes(volume_id)
            if not volume:
                hookenv.log(
                    "Requested volume-id (%s) does not exist. Unable to "
                    "associate storage with %s" % (volume_id, remote_unit),
                    hookenv.ERROR)
                sys.exit(1)

            # Validate that current volume status is supported
            while volume["status"] == "attaching":
                hookenv.log("Volume %s still attaching. Waiting." % volume_id)
                sleep(5)
                volume = self.describe_volumes(volume_id)

            if volume["status"] == "in-use":
                hookenv.log("Volume %s already attached. Done" % volume_id)
                return volume["device"]  # The device path on the instance
            if volume["status"] != "available":
                hookenv.log(
                    "Cannot attach volume. Volume has unsupported status: "
                    "%s" % volume["status"], hookenv.ERROR)
                sys.exit(1)
        else:
            # No volume_id, create a new volume if one isn't already created
            # for the principal of this JUJU_REMOTE_UNIT
            volume_id = self.get_volume_id(volume_label)
            if not volume_id:
                create = getattr(self, "_%s_create_volume" % self.provider)
                if not size:
                    size = hookenv.config("default_volume_size")
                volume_id = create(size, volume_label, instance_id)

        device = None
        hookenv.log("Attaching %s (%s)" % (volume_label, volume_id))
        for x in range(10):
            volume = self.describe_volumes(volume_id)
            if volume["status"] == "in-use":
                return volume["device"]  # The device path on the instance
            if volume["status"] == "available":
                attach = getattr(self, "_%s_attach_volume" % self.provider)
                device = attach(instance_id, volume_id)
                break
            else:
                sleep(5)
        if not device:
            hookenv.log(
                "ERROR: Unable to discover device attached by "
                "euca-attach-volume",
                hookenv.ERROR)
            sys.exit(1)
        return device

    def detach_volume(self, volume_label):
        """Detach a volume from remote unit if present"""
        self.load_environment()   # Will fail if invalid environment
        volume_id = self.get_volume_id(volume_label)

        if volume_id:
            volume = self.describe_volumes(volume_id)
        else:
            hookenv.log("Cannot find volume name to detach, done")
            return

        if volume["status"] == "available":
            hookenv.log("Volume (%s) already detached. Done" % volume_id)
            return

        hookenv.log(
            "Detaching volume (%s) from instance %s" %
            (volume_id, volume["instance_id"]))
        try:
            subprocess.check_call(
                self.commands["detach"] % (volume["instance_id"], volume_id),
                shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log(
                "ERROR: Couldn't detach volume. %s" % str(e), hookenv.ERROR)
            sys.exit(1)
        return

    # EC2-specific methods
    def _ec2_create_tag(self, volume_id, tag_name, tag_value=None):
        """Attach a tag and optional C{tag_value} to the given C{volume_id}"""
        tag_string = tag_name
        if tag_value:
            tag_string += "=%s" % tag_value
        command = 'euca-create-tags %s --tag "%s"' % (volume_id, tag_string)

        try:
            subprocess.check_call(command, shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log(
                "ERROR: Couldn't add tags to the resource. %s" % str(e),
                hookenv.ERROR)
            sys.exit(1)
        hookenv.log("Tagged (%s) to %s." % (tag_string, volume_id))

    def _ec2_describe_instances(self, instance_id=None):
        """
        Use euca2ools libraries to describe instances and return a C{dict}
        """
        result = {}
        try:
            command = self.ec2_instance_class()
            reservations = command.main()
        except SystemExit:
            hookenv.log(
                "ERROR: Couldn't contact EC2 using euca-describe-instances",
                hookenv.ERROR)
            sys.exit(1)
        for reservation in reservations:
            for inst in reservation.instances:
                result[inst.id] = {
                    "ip-address": inst.ip_address, "image-id": inst.image_id,
                    "instance-type": inst.image_id, "kernel": inst.kernel,
                    "private-dns-name": inst.private_dns_name,
                    "public-dns-name": inst.public_dns_name,
                    "reservation-id": reservation.id,
                    "state": inst.state, "tags": inst.tags,
                    "availability_zone": inst.placement}
        if instance_id:
            if instance_id in result:
                return result[instance_id]
            return {}
        return result

    def _ec2_describe_volumes(self, volume_id=None):
        """
        Use euca2ools libraries to describe volumes and return a C{dict}
        """
        result = {}
        try:
            command = self.ec2_volume_class()
            volumes = command.main()
        except SystemExit:
            hookenv.log(
                "ERROR: Couldn't contact EC2 using euca-describe-volumes",
                hookenv.ERROR)
            sys.exit(1)
        for volume in volumes:
            result[volume.id] = {
                "device": "",
                "instance_id": "",
                "size": volume.size,
                "snapshot_id": volume.snapshot_id,
                "status": volume.status,
                "tags": volume.tags,
                "id": volume.id,
                "availability_zone": volume.zone}
            if "volume_name" in volume.tags:
                result[volume.id]["volume_label"] = volume.tags["volume_name"]
            else:
                result[volume.id]["tags"]["volume_name"] = ""
                result[volume.id]["volume_label"] = ""
            if volume.status == "in-use":
                result[volume.id]["instance_id"] = (
                    volume.attach_data.instance_id)
                result[volume.id]["device"] = volume.attach_data.device
        if volume_id:
            if volume_id in result:
                return result[volume_id]
            return {}
        return result

    def _ec2_create_volume(self, size, volume_label, instance_id):
        """Create an EC2 volume with a specific C{size} and C{volume_label}"""
        # Volumes need to be in the same zone as the instance
        hookenv.log(
            "Creating a %sGig volume named (%s) for instance %s" %
            (size, volume_label, instance_id))
        instance = self._ec2_describe_instances(instance_id)
        if not instance:
            config_data = hookenv.config()
            hookenv.log(
                "ERROR: Could not create volume for instance %s. No instance "
                "details discovered by euca-describe-instances. Maybe the "
                "charm configured endpoint %s is not valid for this region." %
                (instance_id, config_data["endpoint"]), hookenv.ERROR)
            sys.exit(1)

        try:
            output = subprocess.check_output(
                "euca-create-volume -z %s -s %s" %
                (instance["availability_zone"], size), shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log("ERROR: %s" % str(e), hookenv.ERROR)
            sys.exit(1)

        response_type, volume_id = output.split()[:2]
        if response_type != "VOLUME":
            hookenv.log(
                "ERROR: Didn't get VOLUME response from euca-create-volume. "
                "Response: %s" % output, hookenv.ERROR)
            sys.exit(1)
        volume = self.describe_volumes(volume_id.strip())
        if not volume:
            hookenv.log(
                "ERROR: Unable to find volume '%s'" % volume_id.strip(),
                hookenv.ERROR)
            sys.exit(1)
        volume_id = volume["id"]
        self._ec2_create_tag(volume_id, "volume_name", volume_label)
        return volume_id

    def _ec2_attach_volume(self, instance_id, volume_id):
        """
        Attach an EC2 C{volume_id} to the provided C{instance_id} and return
        the device path.
        """
        device = "/dev/xvdc"
        try:
            subprocess.check_call(
                "euca-attach-volume -i %s -d %s %s" %
                (instance_id, device, volume_id), shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log("ERROR: %s" % str(e), hookenv.ERROR)
            sys.exit(1)
        return device

    # Nova-specific methods
    def _nova_volume_show(self, volume_id):
        """
        Read detailed information about a C{volume_id} from nova-volume-show
        and return a C{dict} of data we are interested in.
        """
        from ast import literal_eval
        result = {"tags": {}, "instance_id": "", "device": ""}
        command = "nova volume-show '%s'" % volume_id
        try:
            output = subprocess.check_output(command, shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log(
                "ERROR: Failed to get nova volume info. %s" % str(e),
                hookenv.ERROR)
            sys.exit(1)
        for line in output.split("\n"):
            if not line.strip():  # Skip empty lines
                continue
            if "+----" in line or "Property" in line:
                continue
            (_, key, value, _) = line.split("|")
            key = key.strip()
            value = value.replace("null", "None").strip()
            if key in ["availability_zone", "size", "id", "snapshot_id",
                       "status"]:
                result[key] = value
            if key == "display_name":   # added for compatibility with ec2
                result["volume_label"] = value
                result["tags"]["volume_label"] = value
            if key == "attachments":
                attachments = literal_eval(value)
                if attachments:
                    for key, value in attachments[0].items():
                        if key in ["device"]:
                            result[key] = value
                        if key == "server_id":
                            result["instance_id"] = value
        return result

    def _nova_describe_volumes(self, volume_id=None):
        """Create a C{dict} describing all nova volumes"""
        result = {}
        command = "nova volume-list"
        try:
            output = subprocess.check_output(command, shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log("ERROR: %s" % str(e), hookenv.ERROR)
            sys.exit(1)
        for line in output.split("\n"):
            if not line.strip():  # Skip empty lines
                continue
            if "+----" in line or "ID" in line:
                continue
            values = line.split("|")
            if volume_id and values[1].strip() != volume_id:
                continue
            volume = values[1].strip()
            volume_label = values[3].strip()
            if volume_label == "None":
                volume_label = ""
            instance_id = values[6].strip()
            if instance_id == "None":
                instance_id = ""
            result[volume] = {
                "id": volume,
                "tags": {"volume_name": volume_label},
                "status": values[2].strip(),
                "volume_label": volume_label,
                "size": values[4].strip(),
                "instance_id": instance_id}
            if instance_id:
                result[volume].update(self._nova_volume_show(volume))
        if volume_id:
            if volume_id in result:
                return result[volume_id]
            return {}
        return result

    def _nova_attach_volume(self, instance_id, volume_id):
        """
        Attach a Nova C{volume_id} to the provided C{instance_id} and return
        the device path.
        """
        try:
            device = subprocess.check_output(
                "nova volume-attach %s %s auto | egrep -o \"/dev/vd[b-z]\"" %
                (instance_id, volume_id), shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log("ERROR: %s" % str(e), hookenv.ERROR)
            sys.exit(1)
        if device.strip():
            return device.strip()
        return ""

    def _nova_create_volume(self, size, volume_label, instance_id):
        """Create an Nova volume with a specific C{size} and C{volume_label}"""
        hookenv.log(
            "Creating a %sGig volume named (%s) for instance %s" %
            (size, volume_label, instance_id))
        try:
            subprocess.check_call(
                "nova volume-create --display-name '%s' %s" %
                (volume_label, size), shell=True)
        except subprocess.CalledProcessError, e:
            hookenv.log("ERROR: %s" % str(e), hookenv.ERROR)
            sys.exit(1)

        volume_id = self.get_volume_id(volume_label)
        if not volume_id:
            hookenv.log(
                "ERROR: Couldn't find newly created nova volume '%s'." %
                volume_label, hookenv.ERROR)
            sys.exit(1)
        return volume_id


def generate_volume_label(remote_unit):
    """Create a volume label for the requesting remote unit"""
    return "%s unit volume" % remote_unit


def get_running_series():
    lsb_release = "/etc/lsb-release"
    series = {}
    with open(lsb_release) as f:
        for line in f:
            line = line.strip()
            if line.startswith('DISTRIB_CODENAME'):
                series['codename'] = line.split('=')[1]
            if line.startswith('DISTRIB_RELEASE'):
                series['release'] = line.split('=')[1]
            if line.startswith('DISTRIB_DESCRIPTION'):
                series['description'] = line.split('=')[1]
    return series
